#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <cuda_runtime_api.h>
#include <cuda_gl_interop.h>
#include <cutil.h>
#include "cutil_math.h"

#define     TRUE               1
#define     FALSE              0
#define     MATRIX_TRUE        0x1
#define     MATRIX_FALSE       0x0
#define     SHARED_MEM_SZ      (1<<14)   // 16KB

//=========== adjustable settings ==============
//#define     RUB_SHARED_MEM              // #define for RUB with shared memory
#define     RUB_NO_SHARED_MEM           // #define for RUB without shared memory
//#define     RUB_BUFFER_PER_THREAD       // #define for RUB with data block per thread

#define     NUM_PEERS          1024       // # of peers
#define     NUM_BLOCKS         256        // # of buffer blocks, must be < 65535

#define     MAX(a,b)           (((a) >= (b))? (a):(b))
#define     MIN(a,b)           (((a) <= (b))? (a):(b))
#define     iDivUp(a, b)       (((a)+(b)-1)/(b))

// Max number of different blocks to save on shared memory
#define     SMEM_DIFF_BLK_MAX  126        // 98 for 64 blks, 226 for 32 blks
#define     BLK_SZ             64         // 32 or 64
#define     GRID_SZ            iDivUp(NUM_PEERS, BLK_SZ)

#define     GRAPH_PROB         0.5        // Topology: Random Erdos-Renyi Graph with
                                          // parameter p, p=1 means complete graph.
#define     MAX_ROUNDS         (1<<16)
#define     MIN_NEIGHBOUR      100

//==============================================
#if (NUM_BLOCKS > 0xFFFF) // max of USHORT data type
#error "ERROR: NUM_BLOCKS exceeded the limitation"
#endif

#if ((SMEM_DIFF_BLK_MAX * BLK_SZ << 1) > SHARED_MEM_SZ) // max of USHORT data type
#error "ERROR: NUM_BLOCKS exceeded the limitation"
#endif

#define   RANDOM_SELECTION // used for randomly select the peer and blocks
//#define DEBUG_KERNEL
//#define DEBUG_CPU

typedef unsigned int        UINT;
typedef unsigned short      USHORT;
typedef unsigned char       UCHAR;
typedef unsigned char       BOOL;

#include "RUB_kernel.cu"

//============= Functions ======================
void myFree(BOOL **ptr, UINT num)
{
    UINT index;
    if (ptr)
    {
        for(index = 0; index < num; index++)
        {
            if (ptr[index])
                free(ptr[index]);
        }
        free(ptr);
    }
}

void myFree(UINT **ptr, UINT num)
{
    UINT index;
    if (ptr)
    {
        for(index = 0; index < num; index++)
        {
            if (ptr[index])
                free(ptr[index]);
        }
        free(ptr);
    }
}

void myFree(BOOL *ptr)
{
    if (ptr)
        free(ptr);
}

void myFree(USHORT *ptr)
{
    if (ptr)
        free(ptr);
}

void myFree(UINT *ptr)
{
    if (ptr)
        free(ptr);
}

void * myMalloc(size_t size)
{
    void *ptr = malloc(size);
    if (ptr == NULL)
    {
        printf("ERROR: failed to allocated memory\n");
        return(NULL);
    }

    memset(ptr, 0, size);
    return(ptr);
}

UINT myRand(UINT range)
{
    UINT value = rand();
    return(value % range);
}

//Returns a pseudo-random number in (0, 1).
float rand01(void)
{
    float z;
    
    UINT mm = (1 << 30) - 1;
    UINT aa = 16807;
    UINT qq = mm / aa;
    UINT rr = mm % aa;
    UINT seed = rand();
    
    seed = aa * (seed % qq) - rr * (seed / qq);
    if (seed <= 0)
        seed = seed + mm;

    z = (seed * 1.0) / mm;
    if (z > 1.0)
        z -= floor(z);

    return z;
}

void printMatrix(UINT **matrix, UINT num_rows, UINT num_cols)
{
    UINT i, j;
    for(i = 0; i < num_rows; i++)
    {    
        for(j = 0; j < num_cols; j++)
            printf("%d ", matrix[i][j]);

        printf("\n");
    }
    printf("\n");    
}

UINT verifyBufferMatrix(BOOL *matrix, UINT num_rows, UINT num_cols)
{
    UINT i;
    for(i = 0; i < num_rows * num_cols; i++)
    {    
        if (matrix[i] != MATRIX_TRUE)
    {
            printf("WARNING: matrix contain non-finished buffers, row=%d, blk=%d\n", 
                i / num_cols, i % num_cols);
        return(FALSE);
    }
    }

    return(TRUE);
}

void printMatrix(BOOL *matrix, UINT num_rows, UINT num_cols)
{
    UINT i;
    for(i = 0; i < num_rows * num_cols; i++)
    {    
        if (matrix[i] == MATRIX_TRUE)
                printf("1 ");
        else
            printf("0 ");
    
            if ((i+1) % num_cols == 0)
                printf("\n");
    }
    printf("\n");    
}

void printMatrix(UINT *matrix, UINT num_rows, UINT num_cols)
{
    UINT i;
    UINT *ptr = matrix;
    printf("ROW 0:    ");
    for(i = 0; i < num_rows * num_cols; i++, ptr++)
    {    
        printf("%d ", *ptr);

        if ((i+1) % num_cols == 0)
            printf("\nROW %d:    ", (i+1)/num_cols);
    }
    printf("\n");
}


//Generate a random adjacency matrix
void randAdjacency(UINT numPeers, float prob, UINT **peerConnTbl)
{
    UINT  i,j;

    for(i=0; i<numPeers; i++)
    {    
        for(j=i; j<numPeers; j++)
        {
            if( (prob == 1.0 || rand01() < prob) && (i != j) )
            {
                peerConnTbl[i][j] = 1;
                peerConnTbl[j][i] = 1;
            }
            else
            {
                peerConnTbl[i][j] = 0;
                peerConnTbl[j][i] = 0;
            }            
        }
    }
}

//Random/Complete Graph Generator
//Peering information are in peerConnTbl and neighTbl 
//Returns the minimum number of neighbors a peer has.
int randTopoGen(UINT numPeers, float prob, UINT **peerConnTbl, USHORT *neighTbl)
{
    USHORT           i, j;
    USHORT             minNumNeighbor = numPeers - 1;

    //Generate a random adjacency matrix
    randAdjacency(numPeers, prob, peerConnTbl);
    //Generate the Neighbor Information matrix [NeighTbl(i,j)] based on peerConnTbl
    for(i = 0; i < numPeers; i++)
    {
        USHORT l = 1;

        memset(&neighTbl[i*numPeers], 0, sizeof(USHORT) * numPeers);
        for(j = 0; j < numPeers; j++)
        {
            if(peerConnTbl[i][j] == TRUE)
            {
                neighTbl[i*numPeers+l] = j;
                l++;
            }
        }

        neighTbl[i*numPeers] = l-1;
        minNumNeighbor = MIN(minNumNeighbor, neighTbl[i*numPeers]);
    }
    return minNumNeighbor;
}

//Find what is in buffSource, but is not in buffTarget. 
//diff stores the positions (block_ID=i) in the buffermap that buffSource[i]==1 && buffTarget[i]==0
//returns the length of diff.
UINT buffDiff(UINT* diff, BOOL* buffSource, BOOL* buffTarget, UINT length)
{
    UINT i, count=0;
    for(i = 0; i < length; i++)
    {
        if(buffSource[i] == MATRIX_TRUE && buffTarget[i] != MATRIX_TRUE)
        {
            diff[count]=i;
            count++;
        }
    }
    return count;
}

//Initialize the buffer maps. N is the number of peers, k is number of blocks to
//be distributed, B[i] is the pointer to Peer i's buffer map.
//The initial buffer map at each peer (B(i, 1:1:k) is the buffer map of Peer i)
void buffInitialize(UINT numPeers, UINT numBlocks, BOOL *h_bufferMatrix)
{ 
    memset(&h_bufferMatrix[0], MATRIX_TRUE, NUM_BLOCKS * sizeof(BOOL));
    memset(&h_bufferMatrix[NUM_BLOCKS], MATRIX_FALSE, (numPeers-1) * NUM_BLOCKS * sizeof(BOOL));
}

// Copy matrix B1 to B2
void copyMatrix(BOOL **src, BOOL **dest, UINT numRows, UINT numCols)
{
    UINT i, j;
    for (i = 0; i < numRows; i++)
        for (j = 0; j < numCols; j++)
            dest[i][j] = src[i][j];
}

//Protocol 1 Random Useful Block (RUB): In each round, each peer first 
//chooses a random neighbor, and then transmits a random useful block to the receiver.
//If preSetSchedu is 1, use the pre-generated source and target schedules given 
//in int** source and int** target;
//otherwise, generate the source and target schedules on the go and store them
//in int** source and int** target.
UINT p2pRUB(UINT numPeers, UINT numBlocks, BOOL* bufferMatrix, USHORT* h_neighbourList)
{
    UINT   r, i, j;
    USHORT   transmitter, receiver;
    USHORT randNeighIndex;
    UINT   *Set4Trans = NULL;
    UINT   lengthDiff;
    BOOL   *bufMatrixRecv;
    BOOL   finish = FALSE;
    
    // Allocate memory for bufMatrixRecv
    bufMatrixRecv = (BOOL *) myMalloc(numPeers * numBlocks * sizeof(BOOL));
    if (bufMatrixRecv == NULL)
        return(0);

    // Initialize bufMatrixRecv
    memcpy(bufMatrixRecv, bufferMatrix, sizeof(BOOL) * numPeers * numBlocks);
    
    for (r = 0; finish == FALSE && r < MAX_ROUNDS; r++)
    {
        //Round r
        //printf("Round %d\n", r);
        for (i = 0; i < numPeers; i++)
        {            
            transmitter = i;

            // assume every peer will perform in Round r
            //randNeighIndex = (UINT) floor(rand01() * h_neighbourList[i][0]);
            randNeighIndex = myRand(h_neighbourList[i*numPeers]);
            receiver = (UINT) h_neighbourList[i*numPeers+randNeighIndex+1];

      //printf("round %d, peer %d, totalNeighbours %d, randNeighIndex=%d, target=%d\n", r, i, h_neighbourList[i][0], randNeighIndex, receiver;
            
            //Find the blocks that transmitter has and receiver doesn't have. 
            //Store them in Set4Trans, lengthDiff is the length of Set4Trans
            Set4Trans = (UINT *) myMalloc(sizeof(UINT) * numBlocks);
            if (Set4Trans == NULL)
            {
                myFree(bufMatrixRecv);
                return(0);
            }

            lengthDiff = buffDiff(Set4Trans, &bufferMatrix[transmitter*numBlocks], &bufferMatrix[receiver*numBlocks], numBlocks);
            
            //If there is something to trannsmit, then transmit a random block.
            if (lengthDiff > 0)
            {
                //UINT blockID = Set4Trans[(UINT)(rand01() * lengthDiff)];
                UINT blockID = Set4Trans[myRand(lengthDiff)];
                
                //Transmit the block with ID blockID from transmitter to receiver by updating bufMatrixRecv
                //printf("Round %d: Peer %d transmits block %d to Peer %d\n", r, transmitter, blockID, receiver);
                //printf(". ");
                bufMatrixRecv[receiver*numBlocks+blockID] = bufferMatrix[transmitter*numBlocks+blockID];
            }
            free(Set4Trans);
        }

        //update bufferMatrix
        //copyMatrix(bufMatrixRecv, bufferMatrix, numPeers, numBlocks);
        memcpy(bufferMatrix, bufMatrixRecv, sizeof(BOOL) * numPeers * numBlocks);
        
        //If all the peer buffers are filled, then break.
        finish = TRUE;
        for (j = 0; j < numPeers * numBlocks && finish == TRUE; j++)
            if (bufferMatrix[j] != MATRIX_TRUE)
                finish = FALSE;
        
        //printf("\n");
    } // end of Round r

    myFree(bufMatrixRecv);

    return r;
}

/*-----------------------------------------------*/
/*              Main function                    */
/*-----------------------------------------------*/
int main( int argc, char** argv) 
{
    int               dev;
    UINT              timer        = 0;
    float             GPUtotaltime = 0;
    float             CPUtotaltime = 0;
    cudaError_t       d_ret;
    UINT              i;
    UINT              tmp1 = 0;
    UINT              tmp2 = 0;
    UINT              minNumNeighbor;
    UINT              cpuDelayRUB, gpuDelayRUB;
    UINT              **peerConnTable  = NULL;
    USHORT            *h_neighbourList  = NULL;
    BOOL              *h_bufferMatrix   = NULL;

    BOOL              *d_bufferMatrix   = NULL;

    USHORT            *d_neighbourList  = NULL;
    UINT              *d_peerDoneCount;
    UINT              *d_roundCount;
    dim3              blk, grid;

    srand(time(NULL));

    // start timer for CPU calculation
    CUT_DEVICE_INIT(argc, argv);

    /* check GPU device */
    d_ret      = cudaGetDevice(&dev);
    if (d_ret != cudaSuccess)
    {
        printf("CUDA: get device failed\n");
        exit(10);
    }

#ifdef RUB_SHARED_MEM
    printf("============== RUB (with SMEM) ===============\n");
#endif
#ifdef RUB_NO_SHARED_MEM
    printf("============== RUB  (no SMEM)  ===============\n");
#endif
#ifdef RUB_BUFFER_PER_THREAD
    printf("============== RUB (buffer/thread) ===============\n");
#endif
    printf("Allocating CPU memory");
    // Allocate memory for topology matrices and Neighbor
    peerConnTable  = (UINT **) myMalloc(NUM_PEERS * sizeof(UINT *));

    if (peerConnTable == NULL)
        exit(11);
    
    printf(" . . .");
    for(i=0; i<NUM_PEERS; i++)
    {
        peerConnTable[i] = (UINT *) myMalloc(NUM_PEERS * sizeof(UINT));
        if (peerConnTable[i] == NULL)
        {
            myFree(peerConnTable, i);
            exit(12);
        }
    }

    printf(" . . .\n");
    h_neighbourList  = (USHORT *) myMalloc(NUM_PEERS * NUM_PEERS * sizeof(USHORT));
    if (h_neighbourList == NULL)
    {
        myFree(peerConnTable, NUM_PEERS);
        exit(13);
    }
    
    printf("Genrating random graph . . . ");
    // Generate an Erdos-Renyi random graph with N nodes and parameter p
    do
    {
        minNumNeighbor = randTopoGen(NUM_PEERS, GRAPH_PROB, peerConnTable, h_neighbourList);
    } while (minNumNeighbor <= MIN_NEIGHBOUR);

    printf(" min neighbors: %d\n", minNumNeighbor);
    
    // Display the generated topology
    //printMatrix(peerConnTable, NUM_PEERS, NUM_PEERS);
    //printMatrix(h_neighbourList, NUM_PEERS, NUM_PEERS);

    myFree(peerConnTable, NUM_PEERS);

    printf("Allocating memory for buffer matrix . . .\n");
    // Allocate memory for the block matrix, which stores peer buffer states
    // each bit represents one buffer block to save memory usage
    h_bufferMatrix = (BOOL *) myMalloc(NUM_PEERS * NUM_BLOCKS * sizeof(BOOL));
    if (h_bufferMatrix == NULL)
    {
        myFree(h_neighbourList);
        exit(14);
    }

    // Initialize buffer maps
    buffInitialize(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix);
    //printMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);

    printf("Allocating memory on device");

    // Allocate memory on Device and copy data over
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_neighbourList, NUM_PEERS * NUM_PEERS * sizeof(USHORT)));
    if (d_neighbourList == NULL)
    {
        myFree(h_neighbourList);
        myFree(h_bufferMatrix);
        exit(15);
    }
    
    printf(" . . .");
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_bufferMatrix, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL)));
    if (d_bufferMatrix == NULL)
    {
        myFree(h_neighbourList);
        myFree(h_bufferMatrix);
        CUDA_SAFE_CALL(cudaFree(d_neighbourList));
        exit(16);
    }

    printf(" . . .\n");
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_peerDoneCount, sizeof(UINT)));
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_roundCount, sizeof(UINT)));

    if (d_peerDoneCount == NULL || d_roundCount == NULL )
    {
        myFree(h_neighbourList);
        myFree(h_bufferMatrix);
        CUDA_SAFE_CALL(cudaFree(d_neighbourList));
        CUDA_SAFE_CALL(cudaFree(d_bufferMatrix));
        exit(18);
    }

    printf("copying data to device . . .");

    CUDA_SAFE_CALL(cudaMemcpy(d_peerDoneCount, &tmp1, sizeof(UINT), cudaMemcpyHostToDevice ));
    CUDA_SAFE_CALL(cudaMemcpy(d_roundCount, &tmp2, sizeof(UINT), cudaMemcpyHostToDevice ));

    printf(" . . .\n");
    CUDA_SAFE_CALL(cudaMemcpy(d_bufferMatrix, h_bufferMatrix, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL), cudaMemcpyHostToDevice ));
    CUDA_SAFE_CALL(cudaMemcpy(d_neighbourList, h_neighbourList, NUM_PEERS * NUM_PEERS * sizeof(USHORT), cudaMemcpyHostToDevice ));

 #ifdef DEBUG_KERNEL
    printf("buffer matrix source on device\n");
    printMatrix(d_bufferMatrix, NUM_PEERS, NUM_BLOCKS);
 #endif

    blk  = dim3(BLK_SZ, 1, 1);
    grid = dim3(GRID_SZ, 1, 1);

    printf("peer number %d, buffer blocks %d\n", NUM_PEERS, NUM_BLOCKS);
    printf("block size %d, grid size %d\n", BLK_SZ, GRID_SZ);
    //printf("starting GPU RUB ....\n");

    CUT_SAFE_CALL( cutCreateTimer( &timer));

    printf("============== RUB on GPU & CPU  =============\n");
    // start timer for CPU calculation
    cutResetTimer(timer);
    cutStartTimer(timer);

    // run the RUB on GPU
    for(gpuDelayRUB = 0; gpuDelayRUB < MAX_ROUNDS; gpuDelayRUB++)
    {
        #ifdef DEBUG_KERNEL
        printf("============= GPU round %d ==============\n", gpuDelayRUB);
        #endif
#ifdef RUB_SHARED_MEM
        d_p2pRUB_smem <<<grid, blk>>> (d_neighbourList, d_bufferMatrix, d_peerDoneCount, gpuDelayRUB);
#else
#ifdef RUB_NO_SHARED_MEM
        d_p2pRUB_no_smem <<<grid, blk>>> (d_neighbourList, d_bufferMatrix, d_peerDoneCount, gpuDelayRUB);
#else
        d_p2pRUB_bufPerThread <<<grid, blk>>> (d_neighbourList, d_bufferMatrix, d_peerDoneCount, gpuDelayRUB);
#endif
#endif
        CUDA_SAFE_CALL(cudaThreadSynchronize());

        // check if we have finished transmiting
        CUDA_SAFE_CALL(cudaMemcpy(&tmp1, d_peerDoneCount, sizeof(UINT), cudaMemcpyDeviceToHost ));
        if (tmp1 < NUM_PEERS)
        {
            tmp1 = 0;
            CUDA_SAFE_CALL(cudaMemcpy(d_peerDoneCount, &tmp1, sizeof(UINT), cudaMemcpyHostToDevice ));
        }
        else
            break;
    }

    // stop timer
    cutStopTimer(timer);
    GPUtotaltime = cutGetTimerValue(timer);


    CUDA_SAFE_CALL(cudaMemcpy(h_bufferMatrix, d_bufferMatrix, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL), cudaMemcpyDeviceToHost ));
#ifdef DEBUG_CPU
    printf("====== buffer matrix after kernel done =======\n");
    printMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);
#endif
    verifyBufferMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);

    printf("GPU: RUB Rounds = %d, time = %.6f s\n", gpuDelayRUB, GPUtotaltime/1000);

    CUDA_SAFE_CALL(cudaFree(d_neighbourList));
    CUDA_SAFE_CALL(cudaFree(d_bufferMatrix));
    CUDA_SAFE_CALL(cudaFree(d_peerDoneCount));
    //CUDA_SAFE_CALL(cudaFree(d_roundCount));


    //printf("============== RUB on CPU =============\n");
    // Initialize buffer maps
    buffInitialize(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix);
    //printMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);

    // create timer
    cutResetTimer(timer);
    cutStartTimer(timer);

    // Disseminating blocks using RUB
    cpuDelayRUB = p2pRUB(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix, h_neighbourList);

    // stop timer
    cutStopTimer(timer);
    CPUtotaltime = cutGetTimerValue(timer);

    //printMatrix(B, N, k);
    printf("CPU: RUB Rounds = %d, time = %.6f s\n", cpuDelayRUB, CPUtotaltime/1000);
    printf("GPU speedup: %.2f\n", CPUtotaltime/GPUtotaltime);

    //========= memory clean up ==========
    //printf("Deallocating CPU/GPU memory ...\n");
    
    myFree(h_neighbourList);
    myFree(h_bufferMatrix);

    CUT_SAFE_CALL( cutDeleteTimer( timer));

    //check if kernel execution generated and error
    CUT_CHECK_ERROR("Kernel execution failed");
    //CUT_EXIT(argc, argv);

    return(0);
}
